import json

import pypandoc

from autonomous.apis import WikiJS


class MarkdownParser:
    """
    Parses a record into Markdown
    """

    def __init__(self, record):
        self.record = record
        self.markdown = ""
        self.tab = "  "
        self.list_tag = "- "
        self.htag = "#"
        self.listdepth = 0
        self.headerdepth = 0

    def parseJSON(self, json_block):
        if isinstance(json_block, dict):
            self.parseDict(json_block)
        elif isinstance(json_block, list):
            self.parseList(json_block)
        else:
            self.addValue(json_block)

    def parseDict(self, d):
        for k, v in d.items():
            if isinstance(v, (dict, list)):
                self.headerdepth += 1
                self.addHeader(k)
                self.parseJSON(v)
                self.headerdepth -= 1
            else:
                self.addLabeledValue(k, v)
        self.markdown += "\n"

    def parseList(self, l):
        for value in l:
            if isinstance(value, dict):
                self.listdepth = 0
                self.parseDict(value)
            elif isinstance(value, list):
                self.listdepth += 1
                self.parseList(value)
                self.listdepth -= 1
            else:
                self.addListItem(value)

    def addHeader(self, value):
        if self.headerdepth > 6:
            self.headerdepth = 6
        elif self.headerdepth > 1:
            self.markdown += "\n"
        self.listdepth = 0
        self.markdown += f"""{self.htag * (self.headerdepth)} {value}"""
        self.markdown += "\n\n"

    def addListItem(self, value):
        self.markdown += f"""{self.tab * self.listdepth}{self.list_tag}{value}"""
        self.markdown += "\n"

    def addValue(self, value):
        self.markdown += f"{value}"
        self.markdown += "\n"

    def addLabeledValue(self, key, value):
        self.headerdepth += 1
        self.addHeader(key)
        self.headerdepth -= 1
        self.addValue(value)

    def parse(self):
        self.parseJSON(self.record)
        self.markdown = self.markdown.replace("#######", "######")
        return self.markdown


class Page:
    """
    Creates a Page object that can be converted to amrkdown and pushed to a wiki
    Object must be:
        - a dict
        - a class with a __dict__ attribute
        - a class with a serialize() method
    The default wiki is the built in WIkiJS container. Overwrite with your own wiki API by setting the `wiki_api` class attribute.
    Inherit from autonomous.wiki.Wiki abstract class for the required interface.
    """

    wiki_api = WikiJS()
    parser = MarkdownParser

    @classmethod
    def convert(cls, record):
        """
        Converts the object to markdown
        """
        parser = cls.parser(record)
        return parser.parse()

    @classmethod
    def push(cls, record, title, id=None, **kwargs):
        if hasattr(record, "serialize"):
            record = record.serialize()
            record = {title: record}
        elif hasattr(record, "__dict__"):
            record = record.__dict__

        if not isinstance(record, dict):
            raise TypeError(
                "Record must be a dict, have a __dict__ attribute, or have a serialize() method that converts it to a dictionary"
            )

        content = cls.convert(record)

        if id:
            page = cls.wiki_api.update_page(id, content=content, **kwargs)
        else:
            page = cls.wiki_api.create_page(
                title,
                content,
                path=kwargs["path"],
                description=kwargs.get("description", title),
                tags=kwargs.get("tags", []),
            )
        if hasattr(page, "id"):
            return page
        else:
            raise ValueError(f"Page not created. Response: {page}")



# Built-In Modules

# external Modules
from flask import Blueprint, render_template, request, session

from autonomous import log
from autonomous.auth import auth_required

index_page = Blueprint("index", __name__)


@index_page.route("/", methods=("GET",))
def index():
    return render_template("index.html")


@index_page.route(
    "/protected",
    methods=(
        "GET",
        "POST",
    ),
)
@auth_required
def protected():
    return {"result": "success"}


@index_page.route("/add", methods=("POST",))
def add():
    return {"result": "success"}


@index_page.route("/update", methods=("POST",))
def updates():
    return "updated"


@index_page.route("/delete", methods=("POST",))
def delete():
    return "deleted"



import os

import requests

from autonomous.auth import AutoAuth


class GoogleAuth(AutoAuth):
    def __init__(self, state=None):
        self.req_uri = (
            "https://www.googleapis.com/oauth2/v2/userinfo?fields=id,email,name,picture"
        )
        client_id = os.environ.get("GOOGLE_AUTH_CLIENT_ID")
        client_secret = os.environ.get("GOOGLE_AUTH_CLIENT_SECRET")
        redirect_uri = os.environ.get("GOOGLE_AUTH_REDIRECT_URL")

        scope = os.environ.get("GOOGLE_AUTH_SCOPE", ["openid", "email", "profile"])
        if isinstance(scope, str) and "," in scope:
            scope = scope.split(",")
        gduri = os.environ.get("GOOGLE_DISCOVERY_URL")

        response = requests.get(gduri).json()
        authorize_url = response.get("authorization_endpoint")
        token_endpoint = response.get("token_endpoint")
        super().__init__(
            client_id,
            client_secret,
            authorize_url,
            redirect_uri,
            scope,
            token_endpoint,
            state,
        )



import time

import pytest
from tasks import errormocktask, longmocktask, mocktask, parametermocktask

from autonomous import log
from autonomous.tasks import AutoTasks


class TestAutoTasks:
    def test_autotask_connection(self):
        tasks = AutoTasks()
        tasks.clear()
        assert tasks._connection.ping()
        assert tasks.queue
        assert tasks.queue.job_ids == []

    def test_autotask_add(self):
        tasks = AutoTasks()
        tasks.clear()
        job = tasks.task(mocktask)
        assert job.id

    def test_autotask_get(self):
        tasks = AutoTasks()
        tasks.clear()
        job = tasks.task(mocktask)
        log(job.id)
        result = tasks.get_task(job.id)
        assert result

    def test_autotask_status(self):
        tasks = AutoTasks()
        tasks.clear()
        job = tasks.task(mocktask)
        result = job.status
        assert result

    def test_autotask_param(app):
        at = AutoTasks()
        task = at.task(parametermocktask, 1, 2, "hello", key="value")
        assert at.get_task(task.id)
        assert at.get_task(task.id)
        while task.status == "running":
            time.sleep(1)
            log(task.status)

    def test_autotask_result(app):
        at = AutoTasks()
        task = at.task(parametermocktask, 1, 2, "hello", key="value")
        assert at.get_task(task.id)
        while task.status in ["running", "queued"]:
            time.sleep(1)
            log(task.status)
        log(task.status)
        log(task.result)
        log(task.return_value)
        assert task.return_value == 3

    def test_autotask_long(app):
        at = AutoTasks()
        task = at.task(longmocktask)
        time.sleep(2)
        assert at.get_task(task.id)
        while task.status in ["running", "queued", "started"]:
            time.sleep(1)
        assert task.status == "finished"
        assert task.return_value == "success"

    def test_autotask_all(self):
        tasks = AutoTasks()
        tasks.clear()
        (tasks.task(mocktask, 5, i) for i in range(3))
        for t in tasks.get_tasks():
            assert t.return_value or t.status == "running"

    def test_error_task(app):
        at = AutoTasks()
        task = at.task(errormocktask)
        time.sleep(2)
        assert task.status == "failed"
        assert task.return_value == None



from .GHCallbacks import GHRemoteCallbacks

import pygit2
import shutil
import os

class Repository:

    def __init__(self, repo=None, path="./"):
        """
        _summary_

        Args:
            repo (_type_, optional): _description_. Defaults to None.
            path (str, optional): _description_. Defaults to "./".
        """
        self.dir = path
        self.path = f'{path}/{repo.name}'
        if not os.path.exists(self.path):
            self.repo = pygit2.clone_repository(repo.clone_url, self.path, callbacks=GHRemoteCallbacks())
        else:
            repository_path = pygit2.discover_repository(self.path)
            self.repo = pygit2.Repository(repository_path)
            self.pull()

    def copy_file(self, src, dest=None, overwrite=False):
        # f"{doc_dir}/client_update.md", dest=f"planning/sprints/sprint{sprint}"
        """
        copies file from a path to this repository in the given path

        Args:
            src (_type_): _description_
            dest (_type_, optional): _description_. Defaults to None.
            overwrite (bool, optional): _description_. Defaults to False.
        """
        dest_path = self.repo.path.replace(".git/", '')
        if overwrite or not os.path.isfile(f"{dest_path}{dest}"):
            for dir in dest.split('/')[:-1]:
                if not os.path.isdir(f"{dest_path}{dir}"):
                    os.mkdir(f"{dest_path}{dir}")
                dest_path = f"{dest_path}{dir}/"
            shutil.copyfile(src, f"{dest_path}{dest.split('/')[-1]}")

    def pull(self, remote_name='origin', branch='main'):
        """
        _summary_

        Args:
            remote_name (str, optional): _description_. Defaults to 'origin'.
            branch (str, optional): _description_. Defaults to 'main'.

        Raises:
            AssertionError: _description_
            AssertionError: _description_
        """
        for remote in self.repo.remotes:
            remote.fetch(callbacks=GHRemoteCallbacks())
            remote_master_id = self.repo.lookup_reference(f'refs/remotes/origin/{branch}').target
            merge_result, _ = self.repo.merge_analysis(remote_master_id)
            # Up to date, do nothing
            if merge_result & pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE:
                return
            # We can just fastforward
            elif merge_result & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD:
                self.repo.checkout_tree(self.repo.get(remote_master_id))
                try:
                    master_ref = self.repo.lookup_reference(f'refs/heads/{branch}')
                    master_ref.set_target(remote_master_id)
                except KeyError:
                    self.repo.create_branch(branch, self.repo.get(remote_master_id))
                    self.repo.head.set_target(remote_master_id)
            elif merge_result & pygit2.GIT_MERGE_ANALYSIS_NORMAL:
                self.repo.merge(remote_master_id)

                if self.repo.index.conflicts is not None:
                    for conflict in self.repo.index.conflicts:
                        print(f'Conflicts found in: {conflict[0].path}')
                    raise AssertionError('Conflicts, ahhhhh!!')
                user = self.repo.default_signature
                tree = self.repo.index.write_tree()
                target = [self.repo.head.target, remote_master_id]
                self.repo.create_commit('HEAD', user, user, 'Merge!', tree, target)
                # We need to do this or git CLI will think we are still merging.
                self.repo.state_cleanup()
            else:
                raise AssertionError('Unknown merge analysis result')

    def commit(self):
        """
        _summary_
        """
        ref = self.repo.head.name
        parents = [self.repo.head.target]
        index = self.repo.index
        index.add_all()
        index.write()
        author = pygit2.Signature('Steven Moore', 'samoore@binghamton.edu')
        committer = author
        message = "updates from prof"
        tree = index.write_tree()
        self.repo.create_commit(ref, author, committer, message, tree, parents)
        
    def push(self, remote_name='origin', ref=None):
        """
        _summary_

        Args:
            remote_name (str, optional): _description_. Defaults to 'origin'.
            ref (str, optional): _description_. Defaults to 'refs/heads/master:refs/heads/master'.
        """
        if not ref:
            ref = ['refs/heads/main:refs/heads/main']
        for remote in self.repo.remotes:
            remote.push(ref, callbacks=GHRemoteCallbacks())



import multiprocessing
import os
import traceback
from glob import glob

# Non logging stuff
bind = f"{os.environ.get('APP_HOST', '0.0.0.0')}:{os.environ.get('APP_PORT', 80)}"
workers = 2 * multiprocessing.cpu_count() + 1

access_log_format = "ACCESS - %(U)s-%(m)s - res time: %(M)s %(b)s \n"
error_log_format = "ERROR - %(U)s :: \n"
# Access log - records incoming HTTP requests
accesslog = os.getenv("ACCESS_LOG", "-")
# Error log - records Gunicorn server goings-on
errorlog = os.getenv("ERROR_LOG", "-")
loglevel = "info"

# DEVELOPMENT OPTIONS

timeout = 120
workers = 2
capture_output = True  # Whether to send output to the error log
reload = True
reload_extra_files = ["templates/"] + glob("static/**/*", recursive=True)

# def worker_abort(worker):
#     pid = worker.pid
#     print("worker was aborted - {}".format(pid))
#     traceback.print_stack()

# def worker_int(worker):
#     pid = worker.pid
#     print("worker was interrupted - {}".format(pid))
#     traceback.print_stack()

# def worker_exit(server, worker):
#     pid = worker.pid
#     print("worker exited - {}".format(pid))
#     traceback.print_stack()



import os
from base64 import b64decode

import openai

from autonomous import log


class OpenAI:
    def __init__(self):
        openai.api_key = os.environ.get("OPENAI_KEY")

    def generate_image(
        self,
        prompt,
        size="512x512",
        n=1,
    ):
        images = []

        try:
            response = openai.Image.create(
                prompt=prompt, n=n, size=size, response_format="b64_json"
            )
        except Exception as e:
            log(f"{e}\n\n==== Error: fall back to lesser model ====")
            images = ["https://picsum.photos/400/?blur"]
        else:
            for index, image_dict in enumerate(response["data"]):
                image_data = b64decode(image_dict["b64_json"])
                images.append(image_data)
        return images

    def generate_text(self, text, primer_text="", functions=None):
        json_data = {}
        json_data["messages"] = [
            {
                "role": "system",
                "content": primer_text,
            },
            {
                "role": "user",
                "content": text,
            },
        ]

        if isinstance(functions, (list, tuple)):
            json_data.update({"functions": functions})
        elif functions is not None:
            json_data.update({"function_call": {"name": functions["name"]}})
            json_data.update({"functions": [functions]})
        # breakpoint()
        try:
            response = openai.ChatCompletion.create(model="gpt-4-0613", **json_data)
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo-0613", **json_data
            )
        # breakpoint()
        try:
            result = response["choices"][0]["message"]["function_call"]["arguments"]
        except KeyError:
            result = response["choices"][0]["message"]["content"]
        except Exception as e:
            log(f"{type(e)}:{e}\n\n Unable to generate content ====")
            return e

        return result

    def summarize_text(self, text, primer=""):
        message = [
            {
                "role": "system",
                "content": f"You are a highly skilled AI trained in language comprehension and summarization.{primer}",
            },
            {"role": "user", "content": text},
        ]
        try:
            response = openai.ChatCompletion.create(
                model="gpt-4", temperature=0, messages=message
            )
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = openai.ChatCompletion.create(
                model="gpt-4", temperature=0, messages=message
            )
        # breakpoint()

        try:
            result = response["choices"][0]["message"]["content"]
        except Exception as e:
            log(f"{type(e)}:{e}\n\n Unable to generate content ====")
            return e

        return result



import os


#################################################################
#                         CONFIGURATION                         #
#################################################################
class Config:
    APP_NAME = os.environ.get("APP_NAME", __name__)
    HOST = os.environ.get("APP_HOST", "0.0.0.0")
    PORT = os.environ.get("APP_PORT", 5000)
    SECRET_KEY = os.environ.get("SECRET_KEY", "NATASHA")
    DEBUG = os.environ.get("DEBUG", False)
    TESTING = os.environ.get("TESTING", False)
    TRAP_HTTP_EXCEPTIONS = os.environ.get("TRAP_HTTP_EXCEPTIONS", False)



"""
    This module provides a User class that uses the OpenIDAuth class for authentication.
"""
from datetime import datetime

from autonomous import AutoModel, log


class AutoUser(AutoModel):
    """
    This class represents a user who can authenticate using OpenID.
    """

    attributes = {
        "name": "",
        "email": "",
        "last_login": datetime.now(),
        "state": "unauthenticated",
        "token": "",
        "provider": None,
    }

    @classmethod
    def authenticate(cls, user_info, token=None):
        """
        Initiates the authentication process.
        Returns a redirect URL which should be used to redirect the user to the OpenID provider for authentication.
        """
        log(user_info)
        user = AutoUser.find(email=user_info.get("email"))
        if not user:
            user = AutoUser()

        # parse user_info into a user object
        user.name = user_info["name"]
        user.email = user_info["email"]
        user.token = token
        user.last_login = datetime.now()
        user.state = "authenticated"
        user.save()
        return user

    @property
    def is_authenticated(self):
        """
        Returns True if the user is authenticated, False otherwise.
        """
        return self.state == "authenticated"



import urllib.request

import pytest
from app.app import create_app

from autonomous import AutoModel, log
from autonomous.auth.github import GithubAuth
from autonomous.auth.google import GoogleAuth
from autonomous.auth.user import AutoUser


class TestAuth:
    def test_google_auth(self, client):
        auth = GoogleAuth()
        uri, status = auth.authenticate()
        log(uri, status)
        assert uri
        assert status

    def test_github_auth(self):
        auth = GithubAuth()
        uri, status = auth.authenticate()
        log(uri, status)
        assert uri
        assert status

    def test_user(self):
        AutoModel.table().flush_table()
        user_info = {
            "name": "test",
            "email": "test@test.com",
            "token": "testtoken",
        }
        user = AutoUser.authenticate(user_info)
        user2 = AutoUser.authenticate(user_info)
        assert user.pk == user2.pk



# Built-In Modules

# external Modules
from flask import Blueprint, current_app, request, session

from autonomous import log
from autonomous.auth import auth_required

api_page = Blueprint("api", __name__)


@api_page.route("/", methods=("GET",))
def index():
    return {}


@api_page.route("/protected", methods=("GET",))
@auth_required
def protected():
    log(session["user"])
    log(current_app)
    session["user"] = None
    return {**session}



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""
import json
import uuid

from redis.commands.json.path import Path
from redis.commands.search.field import NumericField, TextField
from redis.commands.search.indexDefinition import IndexDefinition, IndexType
from redis.commands.search.query import Query

from autonomous import log

replacements = [
    ",",
    ".",
    "<",
    ">",
    "{",
    "}",
    "[",
    "]",
    '"',
    "'",
    ":",
    ";",
    "!",
    "@",
    "#",
    "$",
    "%",
    "^",
    "&",
    "*",
    "(",
    ")",
    "-",
    "+",
    "=",
    "~",
]


def escape_value(value):
    """
    [summary]
    """
    if isinstance(value, str):
        for v in replacements:
            idx = value.find(v)
            if idx == 0:
                value = value.replace(v, f"\\{v}")
            if idx > 0 and value[idx - 1] != "\\":
                value = value.replace(v, f"\\{v}")
    return value


def deescape_value(value):
    """
    [summary]
    """
    if isinstance(value, str):
        for v in replacements:
            value = value.replace(f"\\{v}", v)
    return value


class Table:
    """
     _summary_

    _extended_summary_
    """

    def __init__(self, name, attributes, db):
        """
        [summary]
        """
        self._db = db
        self.name = name
        self.schema = attributes
        self.index_name = f"idx:{self.name}"

    @property
    def index(self):
        """
        [summary]
        """
        try:
            self._db.ft(self.index_name).info()
        except Exception as e:
            # log(e)
            schema = []
            for attr, data_type in self.schema.items():
                if attr in self.schema:
                    if isinstance(data_type, str):
                        schema.append(
                            TextField(f"$.{attr}", as_name=attr, sortable=True)
                        )
                    elif isinstance(data_type, (int, float)):
                        schema.append(
                            NumericField(f"$.{attr}", as_name=attr, sortable=True)
                        )
                else:
                    log(f"invalid index attribute: {attr}")

            self._db.ft(self.index_name).create_index(
                schema,
                definition=IndexDefinition(
                    prefix=[f"{self.name}:"], index_type=IndexType.JSON
                ),
                temporary=60 * 60 * 24 * 7,  # set indexes to expire in 1 week
            )
        return self._db.ft(self.index_name)

    def save(self, obj):
        """
        [summary]
        """
        # log(obj)
        obj["pk"] = obj.get("pk") or str(uuid.uuid4())
        obj = {k: escape_value(v) for k, v in obj.items()}
        # log(obj, self.count())
        self._db.json().set(f"{self.name}:{obj['pk']}", Path.root_path(), obj)
        obj = {k: deescape_value(v) for k, v in obj.items()}
        return obj["pk"]

    def count(self):
        """
        count _summary_

        _extended_summary_

        :return: _description_
        :rtype: _type_
        """
        result = len(self._db.keys(f"{self.name}:*"))
        return result or 0

    def delete(self, pk):
        """
        [summary]
        """
        pk = escape_value(pk)
        self._db.json().delete(f"{self.name}:{pk}", Path.root_path())

    def find(self, **search_terms):
        """
        find _summary_

        _extended_summary_

        :return: _description_
        :rtype: _type_
        """
        result = self.search(**search_terms)
        return result[0] if result else None

    def search(self, **search_terms):
        """
        Returns an list of objects based on passed arguments
        as key/value pairs
        """

        # log(search_terms, self.index_name, self.index.info())
        matches = []
        for k, v in search_terms.items():
            v = escape_value(v)
            # TODO: figure out how to search nested fields
            if path := search_terms.get("_nested_path"):
                k = f"{path}.{k}"
            query = Query(f"@{k}:{v}")
            results = self.index.search(query)
            matches += [json.loads(d.json) for d in results.docs]
        results = {}
        for obj in matches:
            if obj["pk"] not in results:
                results[obj["pk"]] = {k: deescape_value(v) for k, v in obj.items()}
        return results.values()

    def get(self, pk):
        """
        [summary]
        """
        pk = escape_value(pk)
        try:
            obj = self._db.json().get(f"{self.name}:{pk}", Path.root_path())
        except Exception as e:
            obj = None
            # log(f"no object found with pk:{pk}")
        if obj:
            obj = {k: deescape_value(v) for k, v in obj.items()}
        return obj

    def all(self):
        """
        all _summary_

        _extended_summary_

        :return: _description_
        :rtype: _type_
        """
        keys = self._db.keys(f"{self.name}:*")
        # log(keys)
        objs = self._db.json().mget(keys, Path.root_path()) if keys else []
        for i, obj in enumerate(objs):
            objs[i] = {k: deescape_value(v) for k, v in obj.items()}
        return objs

    def __str__(self):
        return json.dumps(self.all(), indent=4)

    def clear(self):
        """
        clear _summary_

        _extended_summary_
        """
        # breakpoint()
        keys = self._db.keys(f"{self.name}:*")
        return self._db.delete(*keys) if keys else None
